
/* Chrysalide - Outil d'analyse de fichiers binaires
 * args.c - listes d'opérandes rassemblées en arguments
 *
 * Copyright (C) 2010-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "args.h"


#include <malloc.h>


#include "../../operand-int.h"



/* Définition d'un opérande visant une liste d'opérandes Dalvik (instance) */
struct _GDalvikArgsOperand
{
    GArchOperand parent;                    /* Instance parente            */

    GArchOperand **args;                    /* Liste d'arguments           */
    size_t count;                           /* Taille de cette liste       */

};


/* Définition d'un opérande visant une liste d'opérandes Dalvik (classe) */
struct _GDalvikArgsOperandClass
{
    GArchOperandClass parent;               /* Classe parente              */

};


/* Initialise la classe des listes d'opérandes Dalvik. */
static void g_dalvik_args_operand_class_init(GDalvikArgsOperandClass *);

/* Initialise une instance de liste d'opérandes Dalvik. */
static void g_dalvik_args_operand_init(GDalvikArgsOperand *);

/* Supprime toutes les références externes. */
static void g_dalvik_args_operand_dispose(GDalvikArgsOperand *);

/* Procède à la libération totale de la mémoire. */
static void g_dalvik_args_operand_finalize(GDalvikArgsOperand *);

/* Traduit un opérande en version humainement lisible. */
static void g_dalvik_args_operand_print(const GDalvikArgsOperand *, GBufferLine *, AsmSyntax);



/* Indique le type défini par la GLib pour une liste d'arguments Dalvik. */
G_DEFINE_TYPE(GDalvikArgsOperand, g_dalvik_args_operand, G_TYPE_ARCH_OPERAND);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des listes d'opérandes Dalvik.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_class_init(GDalvikArgsOperandClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GArchOperandClass *operand;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);
    operand = G_ARCH_OPERAND_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_dalvik_args_operand_dispose;
    object->finalize = (GObjectFinalizeFunc)g_dalvik_args_operand_finalize;

    operand->print = (operand_print_fc)g_dalvik_args_operand_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance de liste d'opérandes Dalvik.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_init(GDalvikArgsOperand *operand)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_dispose(GDalvikArgsOperand *operand)
{
    size_t i;

    for (i = 0; i < operand->count; i++)
        g_object_unref(G_OBJECT(operand->args[i]));

    G_OBJECT_CLASS(g_dalvik_args_operand_parent_class)->dispose(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_finalize(GDalvikArgsOperand *operand)
{
    G_OBJECT_CLASS(g_dalvik_args_operand_parent_class)->finalize(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Crée un réceptacle pour opérandes Dalvik servant d'arguments.*
*                                                                             *
*  Retour      : Opérande mis en place.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_dalvik_args_operand_new(void)
{
    GDalvikArgsOperand *result;             /* Structure à retourner       */

    result = g_object_new(G_TYPE_DALVIK_ARGS_OPERAND, NULL);

    return G_ARCH_OPERAND(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                line    = ligne tampon où imprimer l'opérande donné.         *
*                syntax  = type de représentation demandée.                   *
*                                                                             *
*  Description : Traduit un opérande en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_print(const GDalvikArgsOperand *operand, GBufferLine *line, AsmSyntax syntax)
{
    size_t i;                               /* Boucle de parcours          */

    g_buffer_line_insert_text(line, BLC_ASSEMBLY, "{", 1, RTT_HOOK);

    if (operand->count > 0)
    {
        g_arch_operand_print(operand->args[0], line, syntax);

        for (i = 1; i < operand->count; i++)
        {
            g_buffer_line_insert_text(line, BLC_ASSEMBLY, ",", 1, RTT_PUNCT);
            g_buffer_line_insert_text(line, BLC_ASSEMBLY, " ", 1, RTT_RAW);

            g_arch_operand_print(operand->args[i], line, syntax);

        }

    }

    g_buffer_line_insert_text(line, BLC_ASSEMBLY, "}", 1, RTT_HOOK);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à compléter.                              *
*                arg     = nouvel argument pour un appel.                     *
*                                                                             *
*  Description : Ajoute un élément à la liste d'arguments Dalvik.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_dalvik_args_operand_add(GDalvikArgsOperand *operand, GArchOperand *arg)
{
    operand->args = (GArchOperand **)realloc(operand->args,
                                             ++operand->count * sizeof(GArchOperand *));

    operand->args[operand->count - 1] = arg;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à compléter.                              *
*                                                                             *
*  Description : Fournit le nombre d'arguments pris en charge.                *
*                                                                             *
*  Retour      : Nombre positif ou nul.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

size_t g_dalvik_args_count(const GDalvikArgsOperand *operand)
{
    return operand->count;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à compléter.                              *
*                index   = indice de l'argument recherché.                    *
*                                                                             *
*  Description : Founit un élément de la liste d'arguments Dalvik.            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_dalvik_args_operand_get(const GDalvikArgsOperand *operand, size_t index)
{
    /* BUG_ON(index >= operand->count) */

    return operand->args[index];

}
